// GLUT drawing code

// by including draw.h, we ensure that the exported prototypes
//   match the function definitions
#include "draw.h"

#include <GL/glut.h>
extern "C" {
#include <OpenGL/glext.h>
}

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <math.h>

using namespace std;

// set shader filenames (unfortunately global, but at least local to this file)
string vsname, fsname;
void setVSname(string filename) {
    vsname = filename;
}
void setFSname(string filename) {
    fsname = filename;
}

// read and load shader update (set*name must have already been called)
extern "C"
void updateShaders(int repeat)
{
    int fd;
    struct stat sb;
    char *sh;
    GLint err;
    static string verr, ferr;

    // read vertex shader (using mmap to map file directly to memory)
    if (vsname.length()) {
	fd = open(vsname.c_str(), O_RDONLY, 0);
	fstat(fd,&sb);
	sh = static_cast<char*>(mmap(0,sb.st_size,PROT_READ,MAP_FILE,fd,0));
	glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
			   sb.st_size, sh);
	munmap(sh,sb.st_size);
	close(fd);

	// report errors, but just once until error message changes
	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB,&err);
	if (err != -1) {
	    string estr = reinterpret_cast<const char *>(
		glGetString(GL_PROGRAM_ERROR_STRING_ARB));
	    if (estr != verr) {
		printf("%s: %s\n",vsname.c_str(), estr.c_str());
		verr = estr;
	    }
	}
	glEnable(GL_VERTEX_PROGRAM_ARB);
    }

    // read fragment shader (same method)
    if (fsname.length()) {
	fd = open(fsname.c_str(), O_RDONLY, 0);
	fstat(fd,&sb);
	sh = static_cast<char*>(mmap(0,sb.st_size,PROT_READ,MAP_FILE,fd,0));
	glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
			   sb.st_size, sh);
	munmap(sh,sb.st_size);
	close(fd);

	// report errors, but just once until error message changes
	glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB,&err);
	if (err != -1) {
	    string estr = reinterpret_cast<const char *>(
		glGetString(GL_PROGRAM_ERROR_STRING_ARB));
	    if (estr != ferr) {
		printf("%s: %s\n",vsname.c_str(), estr.c_str());
		ferr = estr;
	    }
	}

	glEnable(GL_FRAGMENT_PROGRAM_ARB);
    }

    glutPostRedisplay();

    if (repeat) glutTimerFunc(repeat,updateShaders,repeat);
}


// update frame count
void updateFrame(int m)
{
    static int frame[4]={0,0,0,0};
    static int inc[4]={0,0,0,0};
    const int max=500;

    if (m==0) {			// stop changing
	inc[0] = inc[1] = inc[2] = inc[3] = 0;
    }
    else if (m==1) --inc[0];	// red
    else if (m==2) ++inc[0];
    else if (m==3) --inc[1];	// green
    else if (m==4) ++inc[1];
    else if (m==5) --inc[2];	// blue
    else if (m==6) ++inc[2];
    else if (m==7) --inc[3];	// alpha
    else if (m==8) ++inc[3];
    else if (m==9) {		// reset to start
	inc[0] = inc[1] = inc[2] = inc[3] = 0;
	frame[0] = frame[1] = frame[2] = frame[3] = 0;
    }

    float color[4];
    for(int i=0; i<4; ++i) {
	frame[i] += inc[i];
	if (frame[i] < 0)   {inc[i] = 0; frame[i] = 0;}
	if (frame[i] > max) {inc[i] = 0; frame[i] = max;}
	color[i] = (float)frame[i]/max;
    }

    glColor4fv(color);

    if (inc[0] != 0 || inc[1] != 0 || inc[2] != 0 || inc[3] != 0)
	glutPostRedisplay();
}

// this is called every time the screen needs to be redrawn 
extern "C" void 
drawTeapot(void)
{
    // clear old screen contents
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // send current frame
    updateFrame(-1);

    // draw teapot
    glPushMatrix();
    glTranslatef(-10,0,0);
    glutSolidTeapot(90);
    glPopMatrix();
    
    // tell OpenGL to finish drawing and switch the double buffer
    glutSwapBuffers();
}

// this is called every time the screen needs to be redrawn 
extern "C" void 
drawSphere(void)
{
    // clear old screen contents
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // send current frame
    updateFrame(-1);

    // draw sphere
    glPushMatrix();
    glScalef(90,90,90);

    GLUquadric *quadObj = gluNewQuadric();
    gluQuadricDrawStyle(quadObj, GLU_FILL);
    gluQuadricNormals(quadObj, GLU_SMOOTH);
    gluQuadricTexture(quadObj, GL_TRUE);
    gluSphere(quadObj, 1, 40, 20);

    glPopMatrix();
    
    // tell OpenGL to finish drawing and switch the double buffer
    glutSwapBuffers();
}

// this is called every time the screen needs to be redrawn 
extern "C" void 
drawDonut(void)
{
    // clear old screen contents
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // send current frame
    updateFrame(-1);

    // draw torus
    //     glutSolidTorus(45,90,40,20);
    glPushMatrix();
    glScalef(90,90,90);

    // modified from code found in the MESA distribution of GLUT
    const int rings=40, nsides=20;
    const float r=.5, R=1;

    int i, j;
    GLfloat theta, phi, theta1;
    GLfloat cosTheta, sinTheta;
    GLfloat cosTheta1, sinTheta1;
    GLfloat ringDelta, sideDelta;

    ringDelta = 2.0 * M_PI / rings;
    sideDelta = 2.0 * M_PI / nsides;

    theta = 0.0;
    cosTheta = 1.0;
    sinTheta = 0.0;
    for (i = rings - 1; i >= 0; i--) {
	theta1 = theta + ringDelta;
	cosTheta1 = cos(theta1);
	sinTheta1 = sin(theta1);
	glBegin(GL_QUAD_STRIP);
	phi = 0.0;
	for (j = nsides; j >= 0; j--) {
	    GLfloat cosPhi, sinPhi, dist;
	    phi += sideDelta;
	    cosPhi = cos(phi);
	    sinPhi = sin(phi);
	    dist = R + r * cosPhi;

	    glTexCoord2f(theta1/2./M_PI,phi/2./M_PI);
	    glNormal3f(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi);
	    glVertex3f(cosTheta1 * dist, -sinTheta1 * dist, r * sinPhi);
	    glTexCoord2f(theta/2./M_PI,phi/2./M_PI);
	    glNormal3f(cosTheta * cosPhi, -sinTheta * cosPhi, sinPhi);
	    glVertex3f(cosTheta * dist, -sinTheta * dist,  r * sinPhi);
	}
	glEnd();
	theta = theta1;
	cosTheta = cosTheta1;
	sinTheta = sinTheta1;
    }
    glPopMatrix();
    
    // tell OpenGL to finish drawing and switch the double buffer
    glutSwapBuffers();
}

// alternate version that draws a square
extern "C" void 
drawPlane(void)
{
    // clear old screen contents
    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // send current frame
    updateFrame(-1);

    // draw square
    glBegin(GL_QUADS);
    glTexCoord2f(0,0); glNormal3f(0,0,1); glVertex3f(-90,-90,0);
    glTexCoord2f(1,0); glNormal3f(0,0,1); glVertex3f( 90,-90,0);
    glTexCoord2f(1,1); glNormal3f(0,0,1); glVertex3f( 90, 90,0);
    glTexCoord2f(0,1); glNormal3f(0,0,1); glVertex3f(-90, 90,0);
    glPushMatrix();
    glEnd();
    
    // tell OpenGL to finish drawing and switch the double buffer
    glutSwapBuffers();
}
